'' ILI9325 driver using the Touchblade161 design for faster ram to display transfers with a hc137 for group select
'' Average Joe and James Moxham 2012

' notes re groupings of pins
' To shut down a group, deselect it by making P22 low. This makes Y0-Y7 from the 137 all high
' To start a group, the initial state is P0-P20 all HiZ and P22 HiZ (P21 and 23 are audio and there is less audio noise if the pins are not next to each other)
' Passing to a cog, P0-P20 and P22 HiZ ie DIRA all as 0 fo these pins. This is to avoid contention between spin and the cog code
' A routine in Spin or Pasm that changes the 137 will also return P0-P20 and P22 as HiZ.
' In general terms, H is inactive and L is active. So changing a pin from HiZ to H will have no effect as pins all have pullups.
' Any handover to cog code must be done in this HiZ state and all pins in a group need to work with this
 
CON


  _clkmode      = xtal1 + pll16x                        ' use crystal x 16
  _xinfreq      = 5_000_000

  _1ms    = 1_000_000 /     1_000                         ' Divisor for 1 ms

  ' kye sd numbers
  _dopin = 24
  _clkpin = 25
  _dipin = 26
  _cspin = 27
  _cdpin = -1 ' -1 if unused.
  _wppin = -1 ' -1 if unused.
  _rtcres1 = -1 ' -1 always.
  _rtcres2 = -1 ' -1 always.  
  _rtcres3 = -1 ' -1 always.  
  _statuspin = -1 ' Status LED pin number.

' touchscreen pins
  Touch_Clock = 0                        
  'Touch_ChipSelect = done with a 137 select
  Touch_DataIn = 1            
  Touch_DataOut = 2

' audio pins
  _leftChannelVolume = 32
  _rightChannelVolume = 32
  _leftChannelAudioPin = 21 ' -1 ifnot installed.
  _rightChannelAudioPin = 23 ' -1 ifnot installed.

' general purpose constants
  Margin = 4                                            ' some characters have xoffset of -1 or -2 eg v, and won't display on the extreme left edge
  RamDiskStart = 0                                      ' start of the ram disk
OBJ

  fat:           "SD-MMC_FATEngine.spin"                 ' Kye's latest sd driver and no RTC
  spi:           "SPI_ASM"                               ' thanks to Beau - spi driver for touch screen
  str:           "ASCII0_STREngine.spin"                 ' thanks to Kye - string routines
'  term:          "FullDuplexSerial.spin"                ' Thanks to Chip   - for debugging if display not working
  synth :        "tinysynth"                             ' Thanks Ahle2 http://forums.parallax.com/showthread.php?139724-TinySynth-!
 
VAR
   long orientation
   long curx, cury
   long clkcycles               ' for the delay routine
   word ScreenWidth             ' either 240 in portrait or 320 in landscape
   word ScreenHeight            ' 320 in portrait or 240 in landscape
   byte LineOfText1[20]         ' general purpose string buffer
   byte LineOfText2[20]         ' general purpose string buffer
   byte FontHeight              ' size of the current loaded font (pixels = fontsize *.75)
   word BackFontColor           ' the background color in RRRRRGGG GGGBBBBB format
   long RamDiskPointer          ' pointer to current ramdisk location
   long RamDiskFiles            ' number of files in the ramdisk
                  
   'byte sdbuffer[512]     ' can't put these here, writes to sdbuffer sometimes back overwrite variables above this eg fontheight!!
   'byte buffer2[512]      ' so put these in a dat section
   'word rambuffer[256]
   long DecRate                ' synth settings
   long SusLevel
   long PWinit
   long pwmRate
   long Filter
   long Portamento
   
PUB Main | touchtest  ' debug value
  start_ram                                             ' start the external ram / display driver in a cog
  Change161(0)                                          ' reset all the 161 counters to zero
  SelectMemGroup                                        ' select group1
  Start_ILI9325                                         ' start the display. Don't clear screen as this uses ram commands and these might not be working yet
  Text(string("SD"))                         ' string to send, uses internal prop font so can see something if the SD fails to mount   
  SetOrientation(true)                                  '  in portrait (true) or landscape mode (false), must be before clearscreen otherwise don't know screensize   
  Clearscreen(RGBtoWord(0,0,0))                         ' clear screen to black - uses the display driver above 
  fat.fatEngineStart( _dopin, _clkpin, _dipin, _cspin, _wppin, _cdpin, _rtcres1, _rtcres2, _rtcres3)
  fat.mountPartition(0)                                 ' mount the sd card
  RamErase                                              ' erase ram chip fat
  LoadDesktop                                           ' a image 240x320 for the desktop. Load first as there is a ramclear routine that does not delete the first entry
  DrawDesktop
  LoadIcons                                             ' load the mask and icons onto the screen
  ILISetcursor(0,0)                                     ' cursor for debugging to 0,0
  SelectSPIGroup                                        ' selects this group and starts the cog
  Touchscreen                                           ' wait for input

PUB Music | f                                               ' use Lassus font, and use font smoothing and supersampling. 60 points works. Modify chr 45 manually (bar line) as width is 6 but advance is 5
     ILISetcursor(0,0)                                  ' sets curx to  zero and cury to zero
     SetOrientation(false)                              ' http://www.dafont.com/lassus.font
     RamErase1                                          ' erase all but file 0 (desktop)
     Clearscreen(RGBtoWord(255,255,255))                ' white background for music
     f := Loadfont(string("Lass60.ifn"))       'load font into file 1
     'ClearRectangle(curx,cury,screenwidth,screenheight,BackFontColor)  ' clear screen
     TextT(f,string("`",203,"+++8-****-FF/="))          ' line of treble clef music
     ILIcrlf
     TextT(f,string("1",164,"cb-0v-HJKLSSSS-i="))       ' line of bass clef music    
     repeat                                             ' end program


PUB Buttons | ck1x,ck1y,ck1s,ck2x,ck2y,ck2s,rd1x,rd1y,rd1s,rd2x,rd2y,rd2s,rd3x,rd3y,rd3s,bu1x,bu1y,xval,yval,redraw,f,f2 ' test out buttons, radio buttons, checkboxes 
   SetOrientation(true)                                 ' portrait view
   RamErase1                                            ' clear the ram disk 
   clearscreen(RGBtoWord(212,208,200))                  ' background color determined from paintshop
   Textboxclear(10,200,120,215)                         ' clear a text box. height = fontheight+1?
   f := Loadfont(string("MSans14W.ifn"))                ' file 1 black on white font
   curx := 12                                           ' x = x1+2 in textboxclear and see lorem for displaying text from an sd card
   cury := 201                                          ' y = y1 in textboxclear
   TextT(f,string("This is a textbox"))                 ' draw a textbox manually rather than read in a bitmap. Faster
   f2 := Loadfont(string("MSans14.ifn"))                ' file 2 black on gray 14 point seems to be the same as 8.25 in vb.net ? why
   SDBMPtoRam(string("butn8430.bmp"))                   ' file 3 a button 84x30 in size
   SDBMPtoRam(string("Uncheckd.bmp"))                   ' file 4 unchecked checkbox pic3
   SDBMPtoRam(string("Checked.bmp"))                    ' file 5 checked checkbox pic4
   SDBMPtoRam(string("RadioUnc.bmp"))                   ' file 6 unchecked radiobox pic 5
   SDBMPtoRam(string("RadioChk.bmp"))                   ' file 7 checked radiobox pic 6
   SDBMPtoRam(string("hot120.bmp"))                     ' file 8 picturebox demo
   DrawBMPRam(8,115,10)                                 ' draw file 8
   redraw := true    
   ' set all the locations of the objects on the screen
   ck1x := ck2x := rd1x := rd2x := rd3x := 20           ' x all the same 
   ck1y := 10    
   ck2y := ck1y + 40                                    ' y coordinates
   rd1y := ck2y + 40
   rd2y := rd1y + 40
   rd3y := rd2y + 40
   bu1x := 80                                           ' button position
   bu1y := 280
   ck1s := true                                         ' startup status
   ck2s := false
   rd1s := true
   rd2s := false
   rd3s := false
   ' now wait for touch
   SelectSPIGroup                                       ' talk to the spi touchscreen
   repeat
      if redraw                                         ' if redraw true then redraw all the objects
         SelectMemGroup                                 ' display mode
         DrawCheckbox(ck1x,ck1y,ck1s,string("Check1"),4,5,f2)            ' checkbox
         DrawCheckbox(ck2x,ck2y,ck2s,string("Check2"),4,5,f2)            ' checkbox
         DrawRadiobox(rd1x,rd1y,rd1s,string("Radiobutton1"),6,7,f2)         ' radiobox
         DrawRadiobox(rd2x,rd2y,rd2s,string("Radiobutton1"),6,7,f2)         ' radiobox 
         DrawRadiobox(rd3x,rd3y,rd3s,string("Radiobutton2"),6,7,f2)         ' radiobox
         DrawButton(bu1x,bu1y,string("Ok"),3,f2)                  ' button (fixed size at the moment)
         redraw := false
         SelectSPIGroup                                 ' talk to the spi touchscreen  
      yval := TouchYPercent                             ' decode yval 0-100%                
      xval := TouchXPercent                             ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)               ' valid keypress
        xval := (xval * 24) / 10                        ' scale to pixels as can measure from the picture in paintshop
        yval := (yval * 32) / 10                        ' scale to pixels 
        ' cheat a bit here, no need to look for x value as only one column of objects. See calculator for x and y buttons
        case yval
          (ck1y-10)..(ck1y+25): !ck1s                   ' checkbox is small so active area bigger than the checkbox
                                redraw := true          ' redraw the screen
          (ck2y-10)..(ck2y+25): !ck2s                   ' checkbox is small so active area bigger than the checkbox
                                redraw := true         
          (rd1y-10)..(rd1y+25): rd1s :=   true            ' there probably is a smarter way to make one value true and all the rest false
                                rd2s :=   false
                                rd3s :=   false                   
                                redraw := true
          (rd2y-10)..(rd2y+25): rd1s :=   false            
                                rd2s :=   true
                                rd3s :=   false                   
                                redraw := true
          (rd3y-10)..(rd3y+25): rd1s :=   false            
                                rd2s :=   false
                                rd3s :=   true                  
                                redraw := true     
          bu1y..(bu1y + 30): return                     ' button so exit this screen 
        pause1ms(250)                                   ' debounce
   repeat

PUB Keyboard | capslock, backspace,xval,yval,address,ascii,f
   Clearscreen(%00000000_00000000)                      ' clear the screen
   SetOrientation(false)                                ' landscape view
   RamErase1                                            ' clear the ram disk  
   f := Loadfont(string("Arial30.ifn"))                 ' font = file 1
   DrawBMPRam(SDBMPtoRam(string("KeyL.bmp")),0,0)       ' load and draw in one command
   capslock := 0 ' default to off                       ' caps lock 
   backspace := 0 'default to off                       ' backspace is zero
   str.builtstring(true)                                ' clear the working string
   SelectSPIGroup                                       ' talk to the spi touchscreen
   ILISetcursor(margin,27)                              ' cursor to top left
   repeat
      xval := TouchYPercent                           ' decode yval 0-100%  and swap x and y as this is landscape mode              
      yval := TouchXPercent                          ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)               ' valid keypress
        SelectMemGroup                                  ' displaying now
        address :=255
        ascii := 255
        xval := (xval * 32) / 10                        ' scale to pixels as can measure from the picture in paintshop
        yval := 240 - ((yval * 24) / 10)                ' scale to pixels and org top left                        
        case yval
          133..162:address := xval / 32                 ' top row qwertyuiop
          163..188:address := 10 + ((xval-17) / 32)     ' second row addres 10 up, offset
          189..214:address := 19 + ((xval -47) / 32)    ' third row, includes two buttons in one
                   if xval<47
                     address := 255                     ' non printing character capslock
                     !capslock                          ' toggle capslock
                   if xval >273
                     !backspace                         ' backspace character
          215..239:if xval >64 and xval < 256
                      address := 26                     ' space bar 
                   if xval>256
                      address :=27                      ' carriage return (LF added below)   
        if address < 255 and backspace == 0
           ascii := lookupz(address: "Q","W","E","R","T","Y","U","I","O","P","A","S","D","F","G","H","J","K","L","Z","X","C","V","B","N","M",32,13)
          if (capslock == 0) and (ascii => 65) and (ascii =< 90) 
            ascii += 32                                 ' lower case              
          curx := TextChar(f,ascii,curx,cury)              ' only print if a valid ascii character
          str.buildstring(ascii)
          if ascii == 13                                ' add line feed if this is a return
            ascii := 10                                 ' add LF
            curx := TextChar(f,ascii,curx,cury)            ' display line feed
            str.buildstring(ascii)                      ' and add to the built string
        if backspace <>0                                ' backspace redraws the entire string buffer
          backspace := 0                                ' backspace now zero
          str.buildstring(8)                            ' kye's string routine removes a character
          ClearRectangle(0,22,319,132,BackFontColor)  ' clear an area the same color as the font background        
          ILISetcursor(margin,27)                       ' top left of the screen
          TextT(f,str.builtstring(false))        ' redraw text
        
        if xval => 0 and xval <45 and yval => 0 and yval <21
          return                                        ' cancel button
        if xval => 280 and xval < 320 and yval => 0 and yval <21
          result := str.builtstring(true)                   ' the 'send' button returns the string including carriage return, line feeds (can do multiple lines of input)
          return
        pause1ms(250)                                   ' debounce
        SelectSPIGroup                                  ' wait for next keypress



PUB PhotoAlbum | picturenumber
   Clearscreen(%00000000_00000000)                      ' clear the screen
   ILISetCursor(0,0)
   Text(string("Please wait 8s"))  
   RamErase1                                            ' erase all but file 0 (desktop)
   SDBMPtoRam(string("cat.bmp"))                        ' file 1
   SDBMPtoRam(string("wallaby.bmp"))                    ' file 2
   SDBMPtoRam(string("kluger.bmp"))                     ' file 3
   SDBMPtoRam(string("elle.bmp"))                       ' file 4
   SDBMPtoRam(string("balln.bmp"))                      ' file 5
   picturenumber := 1
   repeat until picturenumber > 5
     if picturenumber ==2
        SetOrientation(false)                           ' draw a landscape picture
        DrawBMPram(2,0,0)                                 ' draw picture 2 
        SetOrientation(true)
     else
        DrawBMPRam(Picturenumber,0,0)                   ' draw the picture
     result := Touchswish
     if result > 0
       picturenumber --
     else
       picturenumber ++
   clearscreen(%00000000_00000000)    
   pause1ms(1000) ' otherwise still touching the screen
      
PUB TouchScreen  | yval, xval,pointer
    repeat
      yval := TouchYPercent                               ' decode yval 0-100%               
      xval := TouchXPercent                              ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)
        'SelectMemGroup
        'hex(xval,2)      ' if display any debug messages need to restart the SPI cog as it has been stopped by the debug display
        'hex(yval,2)
        'repeat
        pointer := (yval/20)*4+(xval/25)       ' so can use a case statement 0-19
        if pointer < 20
          SelectMemGroup
        case pointer                                    ' 0 to 19
           0:                   
           1:                   
           2:
           3:
           4:
           5: 
           6:
           7:
           8:  PhotoAlbum
           9:  Calculator                 
           10: LoremPortLand
           11: result := \fat.bootpartition(string("MOVIE.EXE"))     ' reboot and run this program
           12: Keyboard
           13: Music
           14: WavDemo
           15: Buttons
           16: PlaySynthesizer
           17:
           18: 
           19: 

        SetOrientation(true)
        SelectMemGroup                                  ' ready for mem out
        DrawDesktop                                     ' redraw the background
        SelectSPIGroup                                  ' restart the spi driver

PUB LoadIcons                                           ' load all the icons onto the screen. Possibly use a .ini text file for this
   SDBMPtoRam(string("mask.bmp"))                       ' file number 1
   MergeBackgroundIconRam(string("photos.bmp"),0,120)   ' file number 2 and leaves filenumber as 2 ready for the next one
   MergeBackgroundIconRam(string("calc.bmp"),60,120)        ' draw an icon onto the background    
   MergeBackgroundIconRam(string("facebook.bmp"),120,120)   ' draw an icon
   MergeBackgroundIconRam(string("videos.bmp"),180,120)     ' draw an icon
   MergeBackgroundIconRam(string("notes.bmp"),0,180)        ' draw an icon
   MergeBackgroundIconRam(string("music.bmp"),60,180)       ' draw an icon
   MergeBackgroundIconRam(string("ipod.bmp"),120,180)       ' draw an icon 
   MergeBackgroundIconRam(string("prop.bmp"),180,180)       ' prop icon for a GUI buttons, radio, checkboxes etc
   MergeBackgroundIconRam(string("itunes.bmp"),0,240)        ' play some music
   DrawDesktop

PUB DrawDesktop
   DrawBMPram(0,0,0)                                    ' file number zero, location 0,0

PUB LoadDesktop                                         ' first bmp file is not deleted wtih clearramdiskexcludedesktop
   SDBMPtoRam(string("balloon.bmp"))                     ' file number 0

    
PUB PlaySynthesizer | xval,yval,portsw,note ' see next routine for automatic tunes
    synth.start(21, 23)                                 ' start the synth
    portsw := false                                     ' portamento off
    PWinit := 255                                        ' synth settings all 0-255
    pwmRate := 12
    decRate := 200                                      
    susLevel := 255
    filter := 16
    portamento := 6  
    Clearscreen(%00000000_00000000)                     ' 
    RamErase1                                           ' erase all but file 0 (desktop)
    SDBMPtoRam(string("synth2.bmp"))                    ' file 1
    SDBMPtoRam(string("synoff.bmp"))                  ' file 2
    SDBMPtoRam(string("synon.bmp"))                   ' file 3
    SDBMPtoRam(string("synknb.bmp"))                  ' file 4
    SDBMPtoRam(string("synslbk.bmp"))                 ' file 5
    DrawBMPRam(1,0,0)                                   ' draw the picture
    DrawBMPRam(2,148,241)                               ' draw file 2
    DrawBMPRam(3,148,283)                               ' draw file 3
    RefreshSliders                                      ' draw all the sliders                              
    SelectSPIGroup                                        ' talk to the spi touchscreen
    repeat
      yval := TouchYPercent                               ' decode yval 0-100%               
      xval := TouchXPercent                               ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)                    ' valid keypress
         xval := (xval*24)/10                             ' 0-240, use portrait mode but it doesn't matter  
         yval := (yval*32)/10
         if xval>143 and xval <199 and yval >280 and yval <320
           synth.stop                                     ' close down the synthesizer
           return
         if xval >143 and xval <199 and yval > 240 and yval < 275 ' portamento switch
           SelectMemGroup
           !portsw                                      ' change status
           if portsw
                 DrawBMPRam(3,148,241)                  ' switch on 
                 synth.noteOn(30, 19, 60, 12, 3, 6)                             ' turn on a note, next notes slide to this 
           else      
                 DrawBMPRam(2,148,241)                  ' switch off
                 synth.noteOff(18)                                              ' turn off any portamento
           pause1ms(250) ' debounce delay      
           SelectSPIGroup
         if xval >132 and xval <202 and yval >9 and yval <244                   ' one of the 5 sliders
           case yval
             9..9+34:   DecRate := ReadSlider(xval,1)                             ' move the slider and scale value
             53..53+34: SusLevel := ReadSlider(xval,2)     
             88..88+34: PWinit := ReadSlider(xval,3)
             124..124+34:PWMRate := ReadSlider(xval,4)
             162..162+34:Filter := ReadSlider(xval,5)
             210..210+34:Portamento := ReadSlider(xval,6)
         if xval < 53                                                           ' white keys     
          case yval                                     
             0..23 : note := 20                              
             24..56 : note := 22
             57..89: note := 24
             90..119 : note := 25
             120..148: note := 27  
             149..177: note := 29  
             178..207: note := 31  
             208..238: note := 32  
             239..266: note := 34  
             267..297: note := 36  
             298..320: note := 37
         if (xval > 54) and (xval < 118)                     ' black keys a little bit wider so can play them
           case yval
             10..44: note := 21  
             47..77: note := 23
             102..131: note := 26  
             133..162: note := 28  
             166..193: note := 30  
             221..250: note := 33  
             255..284: note := 35
         if note <> 0                                                           ' play note
           if portsw                                              ' portamento switch
              synth.slideToNote(note, portamento/16) 
           else
             synth.pluck(note, decRate/12, susLevel/4, pwinit, pwmRate/6, filter/32)        ' do any scaling here, all values 0-255
           note := 0

PUB SynthSlider(x,y,value)                              ' value is 0-255. assumes bitmaps in ram
    DrawBMPRam(5,x,y)                                   ' draw background
    DrawBMPRam(4,x+((value*100)/450),y+5)               ' draw the knob

PUB RefreshSliders | n
    repeat n from 1 to 6
      RedrawOneSlider(n)

PUB ReadSlider(y,n)
    y := y - 132                                            ' origin of the slider is 132
    y := y*256                                              ' slider is 69 pixels so convert to 256   
    y := y / 69
    if y < 0                                                ' check bounds
      y := 0
    if y > 255
      y := 255
    SelectMemGroup
    RedrawOneSlider(n)                                  ' redraw this slider
    SelectSPIGroup     
    result := y                    

PUB RedrawOneSlider(n)                                  ' n = 1,2,3,4,5,6
   case n
     1:SynthSlider(132,9,decRate)                           ' decay     
     2:SynthSlider(132,53,susLevel)                         ' sustain
     3:SynthSlider(132,88,pwinit)                           ' pwminit
     4:SynthSlider(132,124,pwmRate)                         ' pwmrate
     5:SynthSlider(132,162,filter)                          ' filter (0-12 so multiply by 21)
     6:SynthSlider(132,210,portamento)                       ' portamento 0 - 16
     
' ********************************** end synth routines ************************


PUB Calculator | xval,yval,address,ascii,operand,f
    Clearscreen(%00000000_00000000)                     ' so something happens immediately
    RamErase1                                           ' erase all but file 0 (desktop)
    SDBMPtoRam(string("calcpic.bmp"))                    ' file 1
    DrawBMPRam(1,0,0)                                   ' draw the picture    
    f := LoadFont(string("DSdig.ifn"))                  ' file 2 in the ram disk, use this number when printing fonts
    str.builtstring(true)                               ' clear the working string
    SelectSPIGroup                                      ' talk to the spi touchscreen
    repeat
      yval := TouchYPercent                               ' decode yval 0-100%               
      xval := TouchXPercent                               ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)                    ' valid keypress
        address :=0
        case (yval*32)/10                                    ' convert 0-100 to 0-320 as easier to work with pixels off the picture
          60..101: address :=0                               ' row of a 4x6 matrix
          104..140: address :=4
          146..183: address :=8
          187..222: address :=12
          228..265: address :=16
          271..307: address :=20
        case (xval*24)/10                                   ' column convert percent to 0-239
          0..66: address += 0
          67..119: address +=1
          120..172: address +=2
          173..239: address +=3
        ascii := lookupz(address: "X","Q","Q","Q","C",8,"/","*","7","8","9","-","4","5","6","+","1","2","3","=","0","0",".","=")
        case ascii
          "X"             : return                                              ' close the calculator
          "C"             : str.builtstring(true)                               ' clear the current string
                            CalcClear
          "0".."9",8      : str.buildstring(ascii)                              ' build the string. Can also handle backspace
                            CalcClear                            
                            TextT(f,str.builtstring(false))              ' reprint the string each new char in case a backspace 
          "+","-","*","/" : str.buildstring(ascii) ' display these too so can see typed something
                            CalcClear
                            TextT(f,str.builtstring(false))    
                            str.copy(str.builtstring(true),@lineoftext1)        ' move the string to lineoftext1 and reset
                            operand := ascii
          "=" :             xval := str.decimaltointeger(@lineoftext1)          ' first operand. recycle xval and yval
                            yval := str.decimaltointeger(str.builtstring(true)) ' second operand
                            case operand
                                  "+" : address := xval + yval                  ' add
                                  "-" : address := xval - yval                  ' subtract
                                  "*" : address := xval * yval                  ' multiply
                                  "/" : address := xval / yval                  ' integer division
                            str.str(@lineoftext2,address)                       ' convert answer to string
                            str.trimstring(@lineoftext2) ' if negative, leave the -, if positive remove the leading space
                            CalcClear                                           ' clear the display ready to print the answer
                            TextT(f,@lineoftext2)                               ' print the answer
        SelectSPIGroup                                                          ' restart the spi group if display something
        pause1ms(250)' only display once so delay half a second to "debounce"

PUB CalcClear
     ClearRectangle(25,21,220,48,BackFontColor)       ' clear an area the same color as the font background        
     ILISetcursor(25,15)
  

PUB LoremPortLand
    SetOrientation(false)
    Lorem
    SetOrientation(true)
    Lorem


PUB Lorem | pointer,c,block,f
     RamErase1
     SDtoRam(string("Lorem2.txt"))                      ' file 1, Ramsize[1] is in words, not bytes                                 
     ILISetcursor(0,0)                                  ' sets curx to  zero and cury to zero
     f := Loadfont(string("Times20.ifn"))                    'load font , file 2
     ClearRectangle(curx,cury,screenwidth,screenheight,BackFontColor)
     curx := Margin                                     ' some letters have an xoffset of -1 and so won't display
     repeat block from 0 to 0                           ' block is 128 words = 256 bytes
       docmd("T",@sdbuffer,RamLocation[1]+(block<<7),256)         ' read text into sdbuffer
       pointer :=0
       repeat 256
          c := byte[@sdbuffer][pointer]
         if c=< 126
           curx := TextChar(f,c,curx,cury)                   ' only print if a valid ascii character
         pointer ++
     pause1ms(2000)

CON
 ' *********************************** Ram Driver Routines *******************************

' this is a ram disk that has been optimised for large files that are dumped out to the display
' as such they need to be fast and the file parts need to be contiguous. So normal disk drivers that
' split the file into parts and also allow fragmentation don't work as well here.

' So in order to delete a file and free up space, it is better to load the files in an order that
' means files near the end can be deleted. Deleting a file in the middle of the list means all files
' after this must be deleted. On the other hand, there is no real cost doing this apart from the time
' delay reloading files off the SD card.

' Rountines can be simplified
' 1) Erase all files (simply set FileNumber to zero)
' 2) Erase file by number (can only do the last one so subtract one off FileNumber)
' 3) Erase file by name (no need as can only delete the most recent one, so again just subtract 1)
' 4) Copy a file from SD card and create the file on the ram disk - returns FileNumber +1
' 5) Erase all but the desktop background - Set FileNumber to 1

' routines in this group
' 1) to copy a file from SD to ramdisk - use for all files *except* .bmp files 
' 2) read .bmp files and convert to .ili format so that can dump from the ram disk in one instruction
' 3) dump modified .bmp file from ram disk to display
' 4) erase all files

VAR
   byte FileNumber                                      ' 0 to 15

DAT 
    RamFilenames        byte    $0[256]                 ' store 16 strings of aaaaaaaa.bbb or smaller
    RamLocation         long    $0[15]                  ' 16 file ram locatinos
    RamSize             long    $0[15]                  ' File sizes in words - NOT bytes (divide byte by 2)

PUB RamErase
    Filenumber := 0                                     ' erase lookup table for the ram files, effectively erases the files    

PUB RamErase1                                          ' erase ram but leave the desktop bitmap intact
    Filenumber := 1 

PUB SDtoRam(stringptr) | RamAddress,remainder,sizebytes,sizewords                              ' copy a file from the SD card to the ram disk
    OpenFileRead(stringptr)
    sizebytes := fat.filesize                           ' size in bytes
    sizewords := sizebytes >>1                          ' size in words
    ramaddress := NextLocation                          ' next free place in ram
    RamSize[FileNumber] := sizewords                    ' ram disk works in words, not bytes
    repeat sizebytes >> 9                               ' do blocks of 512 bytes each as natural size of SD blocks
      fat.readdata(@sdbuffer,512)                       ' read from SD card
      docmd("S",@sdbuffer,RamAddress,512)               ' move to ram
      RamAddress += 256                                 ' increment ram address by 256 (half 512 as using words)
    remainder :=  sizebytes & %00000000_00000000_00000001_11111111 ' remainder (if not a multiple of 512) 
    if remainder <> 0
      fat.readdata(@sdbuffer,remainder)
      docmd("S",@sdbuffer,RamAddress,remainder)         ' send out the remainder bytes
    fat.closefile                                       ' close the file
    result := FileNumber                                ' return this file
    FileNumber +=1                                      ' point to next location

PRI NextLocation                                         ' get the next free location in ram based on last files size
    if FileNumber == 0
       RamLocation[0] := RamDiskStart                              ' work out start of this file based on last file size
    else   
       RamLocation[FileNumber] := RamLocation[FileNumber - 1] + RamSize[FileNumber - 1]  ' get location in ram
    result := RamLocation[FileNumber]
    
PUB SDBMPtoRam(stringptr) | width,height,w,i,ramaddress       ' .bmp from sd to ram and convert to 2 byte ili format and reverse order along the way
' http://en.wikipedia.org/wiki/BMP_file_format scroll down about 1/3 of the way for the header formats
' get the width, height
' store these as 2 longs at the beginning of the ram file
' the bitmap format starts at the last row and works up so need to reverse the order
    ramaddress := NextLocation                          ' get the next ram location
    OpenFileRead(stringptr)
    fat.readdata(@sdbuffer,$36)                         ' get the header 0x36 hex bytes
    width := sdbuffer[$12] + sdbuffer[$13] << 8         ' only read in two bytes as never will be >64k wide or high
    height := sdbuffer[$16] + sdbuffer[$17] << 8
    w := width * 3                                      ' this number of bytes per row, and round up to nearest 4 bytes et 327 goes to 328 The size of each row is rounded up to a multiple of 4 bytes (a 32-bit DWORD) by padding.
    w +=3                                               ' add 3 and
    w &= %11111111_11111111_11111111_11111100           ' round down
    docmd("S",@sdbuffer,ramaddress,$36)             ' store header to ram
    i := $36 + ramaddress + ((height - 1) * width)  ' start + header and on the last row and work back
    repeat height
      fat.readdata(@sdbuffer,w)                         ' read in the first row
      docmd("F",@sdbuffer,@rambuffer,width)
      HubToRam(i,width)                                 ' send to ram
      i -= width                                        ' subtract the width, move up the picture
    fat.closefile
    Ramsize[Filenumber] := $36 + (width * height)      ' size of bitmap file in words
    result := FileNumber                                ' returns the current file number
    FileNumber += 1                                     ' add 1 to filenumber

PUB DrawBMPRam(f,x,y) | rampointer,width,height          ' pass the filenumber and the x and y location, searches for this file and if found, displays it
   rampointer := RamLocation[f]                         ' find the start of the file
   docmd("T",@sdbuffer,rampointer,$36)                  ' read in the header
   width := sdbuffer[$12] + sdbuffer[$13] << 8         ' only read in two bytes as never will be >64k wide or high
   height := sdbuffer[$16] + sdbuffer[$17] << 8
   Draw(x,y,x+width-1,y+height-1)                    ' set up the area to draw
   RamToDisplay(RamPointer + $36,width*height)       ' skip the header and display
   result := (width <<16) | (height & $0000FFFF)     ' return the width and height combined into a long

' ***************** end ram disk routines    

PUB MergeBackgroundIconRam(stringptr,x,y) | i,j,k,n                                 ' merge icon and background - if mask =0 then background, else icon (not classic alpha compositing but close enough)   
     ' i is background counter, j is icon counter, k = mask counter
    ' assumes desktop is in ramfile 0, mask in ramfile 1, icon in ramfile 2
    FileNumber := 2 
    SDBMPtoRam(stringptr)                               ' load icon into ram  
    Filenumber := 2                                     ' reset filenumber
    i := RamLocation[0] + $36 + (y*(screenwidth+1)+x)   ' location of picture pixel in ram
    j := RamLocation[2] + $36                           ' location of icon in pixel in ram
    k := RamLocation[1] + $36                           ' location of mask pixel in ram
    repeat 60
      docmd("T",@sdbuffer,i,59)                         ' get 59 pixels from background
      docmd("T",@rambuffer,j,59)                        ' get 59 icon pixels
      docmd("T",@buffer2,k,59)                          ' get words for the mask
      'repeat n from 0 to 58
      '  if (word[@buffer2][n] & %00000000_00011111) > %10000 ' RGB are the same, so use BBBBB, mask 5 low bits, and if > %10000
      '     word[@sdbuffer][n] :=word[@rambuffer][n]     ' if >128 then replace with icon pixel
      docmd("X",@sdbuffer,@rambuffer,@buffer2)          ' pasm version of the above 3 lines    
      docmd("S",@sdbuffer,i,59)                         ' send back modified line to the picture buffer    
      i += screenwidth+1                              ' next line in picture
      j +=59                                            ' icon next line
      k +=59                                            ' next line in mask (bytes) - buffer2     



PUB OpenFileRead(stringptr) ' open a file for reading and display a message if it fails
   result := \fat.openfile(stringptr,"R")
   if fat.partitionerror <> 0                           ' failed to find the file
     Text(stringptr)                         ' print the filename
     Text(result)                            ' print the error message
     repeat                                             ' stop the program

PRI TouchValue | xval,yval,x,y ' returns % values 0-100% (or 255 for no touch)
      xval := TouchX
      yval := TouchY
      if (xval => 500) and (yval =< 3800)                       ' both have to be valid for a touch
        pause1ms(2)                               ' tiny delay then read again as the first reads may not have the correct value
        xval := TouchX
        yval := TouchY
        if (xval => 500) and (yval =< 3800)                 ' and if valid again, now more likely this value is correct
            x := xval - 440                                    ' scale 0-3360
            x /=  32                                     ' scale 0-100
            x <#= 100                                    ' max 100
            x #>=0                                       ' min 0
            x := 100 - x                                 ' so left edge on left
            y := yval -420                               ' scale 0-3180
            y /= 32                                      ' scale 0-100
            y <#= 100                                    ' max 100
            y #>=0                                       ' min 0
            y := 100-y                                   ' so 0 is top left     
            result := (y <<8) | x                       ' return 00000000_00000000_YYYYYYYY_XXXXXXXX   
      if (xval < 500) or (xval >3800)
        result := $0000FFFF                             ' invalid number(s)

PUB TouchXPercent
    result := TouchValue & 255                          ' decode xval 0-100%

PUB TouchYPercent
    result := TouchValue >> 8                           ' decode yval 0-100%  

PUB TouchSwish | xval,oldxval,difference,average ' returns result = +ve or -ve depending on swish direction     ' did finger move left or right?
      SelectSPIGroup ' turn on the spi touchscreen
      oldxval := 255' for startup
      average := 0
      repeat until (average > 30) or (average < -30)
        ' possibly add a decay function here for average
        pause1ms(50)
        xval := TouchXPercent                        ' decode xval 0-100%
        if xval == 255                                  ' not touched
          oldxval := 255                                ' reset oldxval as well
        else
          if oldxval <>255                                ' discard the first reading as the screen is touched
            difference := xval - oldxval
            'printdecimal(xval)
            'printdecimal(difference)
            average := average + difference             ' 
            'printdecimal(average)
            'crlf
          oldxval := xval                                 ' store for next time
      result := average                                 ' return negative or positive value         

PUB TouchStart
      SPI.start(3,0)                                        ' delay,state, start clock high

PRI TouchX
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1101_0000)  ' reads x from  500 to 3700  (off < 500 )
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)

PRI TouchY      
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1001_0000)  ' reads y from 400 to 3800 (off > 3800 ) 
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)  

PUB TextBoxClear(x1,y1,x2,y2)                           ' draw the outline of the box
    ClearRectangle(x1,y1,x2,y2,RGBtoWord(255,255,255))
    Drawline(x1-1,y1-1,x2,y1-1,RGBtoWord(64,64,64))        ' gray outline lines - extract values from paint shop pro
    Drawline(x1-2,y1-2,x2+1,y1-2,RGBtoWord(128,128,128))   ' horizontal upper light gray   
    Drawline(x1-1,y1-1,x1-1,y2,RGBtoWord(64,64,64))         ' vertical left gray
    Drawline(x1-2,y1-2,x1-2,y2+1,RGBtoWord(128,128,128))    ' vertical left light gray
    Drawline(x2+2,y1-2,x2+2,y2+2,RGBtoWord(255,255,255))    ' vertical right white
    Drawline(x1-2,y2+2,x2+2,y2+2,RGBtoWord(255,255,255))   ' horizontal lower white 

PUB DrawCheckbox(x,y,Checked,stringptr,c1,c2,f) | i                ' draw a checkbox x,y,checked, string, file for checked, file for unchecked, font file
   if checked
     DrawBMPRam(c1,x,y) 
   else
     DrawBMPRam(c2,x,y)  
   curx := x+20                                         ' add 20 width
   cury := y
   TextT(f,stringptr)

PUB DrawRadioBox(x,y,Checked,Stringptr,c1,c2,f)| i              ' draw a radio box
   if checked
     DrawBMPRam(c1,x,y)
   else
     DrawBMPRam(c2,x,y) 
   curx := x+20                                         ' add 20 width
   cury := y
   TextT(f,stringptr)
                                           
PUB DrawButton(x,y,Stringptr,c1,f)                ' draw a radio box
   DrawBMPRam(c1,x,y) 
   curx := x + (84/2)-(str.len(stringptr)*4 / 2)        ' centre, avg width 4 per char 84 could be a variable width passed to this routine
   cury := y + 8
   TextT(f,stringptr)     

PUB TextChar(FileN,ascii,x,y) | jump,size,width,height,ramaddress,xoffset,yoffset,xadvance   ' read out a character from a .ifn file stored in ram
  ' pass Fonttable = global
  ' moves to next line if off the end
    ramaddress := RamLocation[FileN]
    jump := ReadRamLong(ramaddress,(ascii << 2) + 256)   ' jump location = ascii *4 plus fonttable + 256
    docmd("T",@buffer2,ramaddress+jump>>1,20)            ' read in the width height as a block = quicker than readramlong
    size := long[@buffer2][0]                           ' size in pixels of this character
    xadvance := long[@buffer2][5]                       ' amount to move to next character   
    'if size > 0 and ascii <> 32                         ' no need to print anything if size is zero or read more font data
    width := long[@buffer2][1]                        ' width
    height := long[@buffer2][2]                       ' height
    xoffset := long[@buffer2][3]                      ' xoffset to move
    yoffset := long[@buffer2][4]                      ' yoffset to move
    if (x+width -1 + xoffset) > screenwidth
        ILIcrlf                                         ' do a new line if it won't fit
        x := curx
        y := cury
    if ascii < 32
      xadvance := 0                                     ' add nothing if a non displaying character
    if ascii == 13                                  ' carriage return
        ILIcr
        x := curx
    if ascii ==10                                   ' line feed (new line)
        ILIlf
        y := cury
    if ascii > 32 ' space not displayed
      Draw(x+xoffset,y+yoffset,x+width-1+xoffset,y+height-1+yoffset)      ' draw on screen
      RamToDisplay(ramaddress+((jump+32)>>1),size)       ' move bytes from ram out to the display
    return x + xadvance    

PUB ILIcrlf
    ILIcr
    ILIlf

PUB ILIcr                                               ' carriage return, to left margin
    curx := Margin                                      ' some characters are -1 xoffset so won't display

PUB ILIlf                                               ' line feed, move up one
    cury += FontHeight
    if cury > screenheight
      ClearRectangle(0,0,screenwidth,screenheight,BackFontColor) 
      cury := 0                                         ' if off the bottom of the screen then clear screen start at the top (VT00 does scrolling)
    ' don't clear background here, do in calling routine
    'ClearFontBackground(0,cury,screenwidth,cury + FontHeight) ' clear background line to the background color ready to draw          

PUB TextT(FileN,stringptr)                 'print at curx,cury ising file number in ramdisk
'' Print a zero-terminated string
  repeat strsize(stringptr)
    curx := TextChar(FileN,byte[stringptr++],curx,cury)    


PUB ILISetCursor(x,y)            
    curx := x
    cury := y


PUB ReadRamLong(WordStart,Bytevalue)                   ' easier calculating
    docmd("T",@result,Wordstart+Bytevalue>>1,2)
    return result

PUB LoadFont(stringptr) | filesize,i                ' read a .ifn file premade in angelcode bmfont and then the vb.net program
    SDtoRam(stringptr)                                  ' font to ram disk, adds one to filenumber
    docmd("T",@rambuffer,RamLocation[Filenumber-1],128) ' read first 128 words =256 bytes back
    FontHeight := byte[@rambuffer][251]                 ' font height
    repeat i from 0 to 2
      sdbuffer[i] := byte[@rambuffer][35+i]
    docmd("E",@sdbuffer,@buffer2,1)                     ' convert to 2 byte .ili format
    BackFontColor := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order
    result := FileNumber-1                              ' return pointer to the font file in ram

PUB HubToRam(RamAddress,Number)                         ' send pixels from rambuffer hub to ram
   docmd("S",@rambuffer,RamAddress,Number)

PUB RamToHub(RamAddress,Number)                         ' send pixels from ram to hub at rambuffer
   docmd("T",@rambuffer,RamAddress,Number)

PUB RamToDisplay(RamAddress,Number)                     ' send pixels from ram to display
   docmd("U",0,RamAddress,Number)

PUB HubToDisplay(HubAddress,Number)                     ' send pixels from hub to display
   docmd("V",HubAddress,0,Number)

PUB Change137(group)
   docmd("Y",group,0,0)                                 ' change the current group

PUB Change161(RamAddress)                               ' pass ramaddress. Returns in group 1
   docmd("Z",0,RamAddress,0)   

PUB Clearscreen(color)                                  ' uses 2 byte ILI color (see below for rgb conversion)
    wordfill(@rambuffer,color,256)
    Draw(0,0,screenwidth,screenheight)                           ' set up the area of the screen to draw in   
    repeat 300
      HubToDisplay(@rambuffer,256)

PUB RGBtoWord(R,G,B)                                     ' convert RGB to 2 byte ILI color 
    sdbuffer[0] := R
    sdbuffer[1] := G
    sdbuffer[2] := B                        
    docmd("E",@sdbuffer,@buffer2,1)
    result := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order 
    
PUB ClearRectangle(x1,y1,x2,y2,ColorWord) | i,size,remainder ' clears an area of the screen to the current font background color
    size := (x2-x1+1) * (y2 - y1 +1)                    ' number of pixels
    remainder := size & 255                             ' if not a whole number of 256 pixels
    repeat i from 0 to 255                              ' move the background font colour to the buffer
      rambuffer[i] := ColorWord
    Draw(x1,y1,x2,y2)                                   ' set up the area of the screen to draw in
    repeat size >> 8                                    ' 256 pixel blocks
      docmd("V",@rambuffer,0,256)                
    if remainder <> 0  
      docmd("V",@rambuffer,0,remainder)                 ' do the remainder

PUB Drawline(x1,y1,x2,y2,ColorWord)                     ' draws a vertical or horizontal line
   ClearRectangle(x1,y1,x2,y2,ColorWord)                ' same as drawing a rectangle 
      
PUB SetOrientation(e) ' change orientation true = portrait. Changes global variable 'orientation' and also screen width and height
   orientation := e
   ChangeOrientation(orientation)
   if orientation
     screenwidth :=239
     screenheight :=319
   else
     screenwidth :=319
     screenheight :=239

PUB ChangeOrientation(n) ' pass true = portrait or false = landscape, changes global variable orientation in this object
' command $0001
'8.2.4. Driver Output Control (R01h)
'R/W RS D15 D14 D13 D12 D11 D10 D9 D8 D7 D6 D5 D4 D3 D2 D1 D0
'W 1 0 0 0 0 0 SM 0 SS 0 0 0 0 0 0 0 0
'SS: Select the shift direction of outputs from the source driver.
'When SS = 0, the shift direction of outputs is from S1 to S720
'When SS = 1, the shift direction of outputs is from S720 to S1.
'In addition to the shift direction, the settings for both SS and BGR bits are required to change the
'assignment of R, G, B dots to the source driver pins.
'To assign R, G, B dots to the source driver pins from S1 to S720, set SS = 0.
'To assign R, G, B dots to the source driver pins from S720 to S1, set SS = 1.

' command $0003
 'R/W RS D15 D14 D13 D12 D11 D10 D9  D8 D7  D6   D5   D4   D3 D2 D1 D0
   'W   1  TRI DFM 0   BGR 0   0   HWM 0  ORG 0    I/D1 I/D0 AM 0  0  0
   ' When AM = 0, the address is updated in horizontal writing direction.
   'When AM = 1, the address is updated in vertical writing direction.
   'I/D[1:0] Control the address counter (AC) to automatically increase or decrease by 1 when update one pixel
   ' display data.
    SelectMemGroup                                      ' 137 and appropriate pins
    orientation := n
    if orientation
      ' for origin top left and portrait mode
      ILIcmd($0001,%00000001_00000000)                               ' $0001,$0100 set SS and SM bit 0001 0100   portrait
      ILIcmd($0003,%00010000_00110000)                               ' $0003,$1030 set GRAM write direction and BGR=1. $0003 $1030  
    else
      ' for origin top right and landscape mode 
      ILIcmd($0001,%00000000_00000000)                               ' set SS and SM bit 0001 0000   landscape   
      ILIcmd($0003,%00010000_00111000)                  ' landscape  $1028 = original but 1038 is correct - not mirror image



PUB SelectMemGroup                                      ' select group 2 for memory transfer
   spi.stop                                             ' and stop any other cogs here too if needed
   
   OUTA |= %00000000_00011111_11111111_11111111         ' set P0-P20 high 
   DIRA |= %00000000_00011111_11111111_11111111         ' set P0-P20 as outputs 
   Change137(1)
   OUTA |= %00000000_00011111_00000000_00000000         ' set P16-P20 high 
   DIRA |= %00000000_00011111_11111111_11111111         ' set P0-P20 as outputs 
   
PUB SelectSPIGroup                                      ' select group 0 for the SPI
   OUTA |= %00000000_00000000_00000010_00000000   ' keep p9 high so doesn't reset
   DIRA |= %00000000_00000000_00000010_00000000   ' P0-P2 are used to load the 137
   Change137(2)
   OUTA |= %00000000_00000000_00000010_00000000   ' keep p9 high so doesn't reset
   DIRA |= %00000000_00000000_00000010_00000000   ' enable these pins for output
   DIRA &= %11111111_11111111_11111111_11111000   ' so no clashes with the cog using P0-P2 set these as inputs      
   TouchStart                                     ' start the touchscreen cog

PUB Text(stringptr)                          'print at curx,cury
  repeat strsize(stringptr)
    Propfont_out(byte[stringptr++])
  crlf

PUB crlf
    curx := 0
    cury += 32                                          ' new line at end of string
    if cury >319                                        ' bottom of screen so new screen
      curx:=0
      cury:=0

PUB Propfont_out(ascii) | address,pixels
    Draw(curx,cury,curx+15,cury+31)                     ' location to start drawing 
    address := $8000 + (ascii >> 1) << 7                ' get rom address
    repeat 32                                           ' 32 rows per character, split in two parts
      pixels := long[address]                           ' get rom font data
      pixels := pixels >> (ascii & 1)                   ' shift for odd characters
      repeat 16 ' 16 columns
        if pixels & 1
          Pixel(%00000111_11100000)                     ' foreground color RRRRRGGG_GGGBBBBB
        else
          Pixel(%00000000_00000000)                     ' background color 
        pixels := pixels >> 2                           ' alternate pixels interleaved so shift 2
      address += 4 
    curx +=16
    if curx >239                                        ' new line
      crlf

PUB Start_ILI9325 ' pass orientation true = portrait, false = landscape
    ILI_Reset_High
    pause1ms(5)
    ILI_Reset_Low
    pause1ms(5)
    ILI_Reset_High                                      ' returns in group 2, ie CS is high
    Change137(2)                                        ' replaces ILI_CS high - select another group
    ILI_RD_High
    ILI_WR_High
    pause1ms(5)

    'ILI_CS_Low
    SelectMemGroup                                      ' enables correct pins for output and sets CS low

          ' ************* Start Initial Sequence **********
    ILIcmd($00E5,$78F0)                                 ' set SRAM internal timing
    ILIcmd($0001,$0100)                                 ' set SS and SM bit 0001 0100   portrait
    ILIcmd($0002,$0700)                                 ' set 1 line inversion 
    ILIcmd($0003,$1030)                                 ' set GRAM write direction and BGR=1. $0003 $1030
    ILIcmd($0004,$0000)                                 ' Resize register 
    ILIcmd($0008,$0207)                                 ' set the back porch and front porch 
    ILIcmd($0009,$0000)                                 ' set non-display area refresh cycle ISC[3:0]  
    ILIcmd($000A,$0000)                                 ' FMARK function  
    ILIcmd($000C,$0000)                                 ' RGB interface setting   
    ILIcmd($000D,$0000)                                 ' Frame marker Position   
    ILIcmd($000F,$0000)                                 ' RGB interface polarity  
'        *************Power On sequence ****************//
    ILIcmd($0010,$0000)                                 ' SAP, BT[3:0], AP, DSTB, SLP, STB 
    ILIcmd($0011,$0007)                                 ' DC1[2:0], DC0[2:0], VC[2:0] 
    ILIcmd($0012,$0000)                                 ' VREG1OUT voltage   
    ILIcmd($0013,$0000)                                 ' VDV[4:0] for VCOM amplitude  
    ILIcmd($0007,$0001)   
    pause1ms(50)                                        ' Dis-charge capacitor power voltage
    ILIcmd($0010,$1090)                                 ' 1490//SAP, BT[3:0], AP, DSTB, SLP, STB   
    ILIcmd($0011,$0227)                                 ' DC1[2:0], DC0[2:0], VC[2:0] 
    pause1ms(50)                                        ' delay
    ILIcmd($0012,$001F)                                 ' 001C// Internal reference voltage= Vci;
    pause1ms(50)                                        ' delay   
    ILIcmd($0013,$1500)                                 ' $1000//1400   Set VDV[4:0] for VCOM amplitude  1A00  
    ILIcmd($0029,$0027)                                 ' $0012 //001a  Set VCM[5:0] for VCOMH  //$0025  0034  
    ILIcmd($002B,$000D)                                 ' Set Frame Rate   000C       
    pause1ms(50)                                        ' delay   
    ILIcmd($0020,$0000)                                 ' GRAM horizontal Address   
    ILIcmd($0021,$0000)                                 ' GRAM Vertical Address    
'         ----------- Adjust the Gamma Curve ----------//
    ILIcmd($0030,$0000) 
    ILIcmd($0031,$0707)  
    ILIcmd($0032,$0307)  
    ILIcmd($0035,$0200)  
    ILIcmd($0036,$0008)  
    ILIcmd($0037,$0004)  
    ILIcmd($0038,$0000)  
    ILIcmd($0039,$0707)  
    ILIcmd($003C,$0002)  
    ILIcmd($003D,$1D04)  
'        ------------------ Set GRAM area ---------------//
    ILIcmd($0050,$0000)                                 ' Horizontal GRAM Start Address
    ILIcmd($0051,$00EF)                                 ' Horizontal GRAM End Address       
    ILIcmd($0052,$0000)                                 ' Vertical GRAM Start Address    
    ILIcmd($0053,$013F)                                 ' Vertical GRAM Start Address       
    ILIcmd($0060,$A700)                                 ' Gate Scan Line     
    ILIcmd($0061,$0001)                                 ' NDL,VLE, REV    
    ILIcmd($006A,$0000)                                 ' set scrolling line      
'        -------------- Partial Display Control ---------/
    ILIcmd($0080,$0000)
    ILIcmd($0081,$0000)
    ILIcmd($0082,$0000)
    ILIcmd($0083,$0000)
    ILIcmd($0084,$0000)
    ILIcmd($0085,$0000)
'        //-------------- Panel Control -------------------//
    ILIcmd($0090,$0010)
    ILIcmd($0092,$0600)
    ILIcmd($0007,$0133)                                 ' 262K color and display ON
    ChangeOrientation(true)                             ' default to portrait

PUB Draw(x1, y1, x2, y2)                                ' sets the pixel to x1,y1 and then fills the next (x2-x1)*(y2-y1) pixels
    SelectMemGroup                                      ' set 137 to correct value
    ifnot orientation                                   ' landscape mode so swap x and y
      result :=x1                                       ' swap x1 and y1
      x1 := y1
      y1 := result
      result := x2                                      ' swap x2 and y2
      x2 :=y2
      y2 := result
    ILIcmd($0050,x1)
    ILIcmd($0052,y1)
    ILIcmd($0051,x2)
    ILIcmd($0053,y2)
    ILIcmd($0020,x1) 
    ILIcmd($0021,y1)
    Lcd_Write_Com($0022)


PUB Pixel(pixelcolor)               ' send out a pixel
    Lcd_Write_Data(pixelcolor)

PUB pause1ms(period)

'' Pause execution for period (in units of 1 ms).

  clkcycles := ((clkfreq / _1ms * period) - 4296) #> 381     ' Calculate 1 ms time unit
  waitcnt(clkcycles + cnt)                                   ' Wait for designated time

PUB hex(value, digits)   ' uses internal prop font and slow pixel drawing (for very basic debugging without an SD font)
  SelectMemGroup                               ' reselect 138 for display output for debugging  
'' Print a hexadecimal number
  propfont_out("O")
  propfont_out("x")
  value <<= (8 - digits) << 2
  repeat digits
    propfont_out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))
  crlf
                   

PRI ILIcmd(c,d)  ' instruction in one method
    Lcd_Write_Com(c) ' send out a word
    Lcd_Write_Data(d)

PRI Lcd_Write_Com(ILIlong)
    ILI_RS_low
    LCD_Writ_Bus(ILIlong)
    'docmd("W",ILIlong,0,0)  ' try moving to pasm, leaves spin and cog dira as 0 so need to reset both
    'SelectMemGroup           ' not working though. Not sure if should be ILIlong or @ILIlong. neither work
    

PRI Lcd_Write_Data(ILIlong)
    ILI_RS_High
    LCD_Writ_Bus(ILIlong)    

PRI LCD_Writ_Bus(ILILong)
    'ILILong &= %00000000_00000000_11111111_11111111    ' mask to a word. Not needed if care taken always to send a word value
    OUTA &= %11111111_11111111_00000000_00000000        ' set P0-P15 to zero ready to OR
    OUTA |= ILILong                                     ' merge with the word to output
    ILI_WR_Low                                          ' send out the data
    ILI_WR_High

PRI ILI_RS_Low
   OUTA &= %11111111_11111011_11111111_11111111         ' P18 low

PRI ILI_RS_High   
   OUTA |= %00000000_00000100_00000000_00000000         ' P18 high

PRI ILI_WR_Low
   OUTA &= %11111111_11101111_11111111_11111111         ' p20 low

PRI ILI_WR_High
   OUTA |= %00000000_00010000_00000000_00000000         ' p20 high

PRI ILI_RD_High                                         ' not used, set high in hardware     

'PRI ILI_CS_Low                                          
'   Change137(1)                                          ' change to group 1 
'
'PRI ILI_CS_High
'   Change137(0)                                         ' change to group 0 (any group other than 1)
   

PRI ILI_RESET_Low                                       ' reset low
   Change137(2)                                         ' group 2
   DIRA |= %00000000_00000000_00000010_00000000         ' set pin 9 as an output
   OUTA &= %11111111_11111111_11111101_11111111         ' set pin 9 low

PRI ILI_RESET_High
   Change137(2)                                         ' group 2
   DIRA |= %00000000_00000000_00000010_00000000         ' set pin 9 as an output
   OUTA |= %00000000_00000000_00000010_00000000         ' set pin 9 high

DAT
sdbuffer        byte    $0[1024]  ' 1024 byte buffer for sd card interface and also for sending one row 320 x3 = 960 bytes max picture size 
buffer2         byte    $0[512]   ' 512 general purpose hub buffer 
rambuffer       word    $0[256]   ' 512 byte (256 word) buffer easier for using with words

PUB start_ram : err_
' Initialise the Drac Ram driver. No actual changes to ram as the read/write routines handle this
  command := "I"
  cog := 1 + cognew(@tbp2_start, @command)
  if cog == 0
    err_ := $FF                 ' error = no cog
  else
    repeat while command        ' driver cog sets =0 when done
    err_ := errx                ' driver cog sets =0 if no error, else xx = error code

PUB stop_ram
   if cog
      cogstop(cog~ - 1)      

PUB DoCmd(command_, hub_address, ram_address, block_length) : err_| a,b
' Do the command: A-Z (I is reserved for Initialise)
  'dira := 0                     ' tristate all pins with the spin dira
  a := dira                     ' store the state of these and restore at the end 
  b := outa                     
  DIRA &= %11111111_10100000_00000000_00000000 ' tristate all common pins that cog can change so no conflicts
  hubaddrs := hub_address       ' hub address start
  ramaddrs := ram_address       ' ram address start
  blocklen := block_length      ' block length
  command  := command_          ' must be last !!
' Wait for command to complete and get status
  repeat while command          ' driver cog sets =0 when done
  err_ := errx                  ' driver cog sets =0 if no error, else xx = error code
  dira := a                     ' restore dira and outa
  outa := b

CON
'' Modified code from Cluso's triblade
'' commands to move blocks of data to the ILI9325 touchscreen display
' DoCmd(command_, hub_address, ram_address, block_length)
' I - initialise     
' S - Move data from hub to ram
' T - Move data from ram to hub
' U - Move data from ram to display
' V - Hub to display
' W - not working - writecom in pasm
' E - convert from .raw RGB to two byte ILI format RRRRRGGG_GGG_BBBBB
' F - convert from .bmp BGR format to two byte ILI format
' X - merge icon and background based on a mask
' Y - Change 137 output Returns P0-P20 and P22 in HiZ. Pass hubaddrs
' Z - Set 161 pins. Returns in group 1

VAR

' communication params(5) between cog driver code - only "command" and "errx" are modified by the driver
   long  command, hubaddrs, ramaddrs, blocklen, errx, cog ' rendezvous between spin and assembly (can be used cog to cog)
'        command  = A to Z etc =0 when operation completed by cog
'        hubaddrs = hub address for data buffer
'        ramaddrs = ram address for data
'        blocklen = ram buffer length for data transfer
'        errx     = returns =0 (false=good), else <>0 (true & error code)
'        cog      = cog no of driver (set by spin start routine)
   

DAT
'' +-----------------------------------------------------------------------------------------------+
'' | Touchblade 161 Ram Driver (with grateful acknowlegements to Cluso and Average Joe)            |
'' +-----------------------------------------------------------------------------------------------+
                        org     0
tbp2_start    ' setup the pointers to the hub command interface (saves execution time later
                                      '  +-- These instructions are overwritten as variables after start
comptr                  mov     comptr, par     ' -|  hub pointer to command                
hubptr                  mov     hubptr, par     '  |  hub pointer to hub address            
ramptr                  add     hubptr, #4      '  |  hub pointer to ram address            
lenptr                  mov     ramptr, par     '  |  hub pointer to length                 
errptr                  add     ramptr, #8      '  |  hub pointer to error status           
cmd                     mov     lenptr, par     '  |  command  I/R/W/G/P/Q                  
hubaddr                 add     lenptr, #12     '  |  hub address                           
ramaddr                 mov     errptr, par     '  |  ram address                           
len                     add     errptr, #16     '  |  length                                
err                     nop                     ' -+  error status returned (=0=false=good) 


' Initialise hardware tristates everything and read/write set the pins
init                    mov     err, #0                  ' reset err=false=good
                        'mov     dira,zero                ' tristate the pins with the cog dira
                        and     dira,maskP0P20P22       ' tristates all the common pins

done                    wrlong  err, errptr             ' status  =0=false=good, else error x
                        wrlong  zero, comptr            ' command =0 (done)
' wait for a command (pause short time to reduce power)
pause
'                        mov     ctr, delay      wz      ' if =0 no pause
'              if_nz     add     ctr, cnt
'              if_nz     waitcnt ctr, #0                 ' wait for a short time (reduces power)
                        rdlong  cmd, comptr     wz      ' command ?
              if_z      jmp     #pause                  ' not yet
' decode command
                        cmp     cmd, #"S"       wz      ' hub to ram
              if_z      jmp     #pasmhubtoram           
                        cmp     cmd, #"T"       wz      ' ram to hub
              if_z      jmp     #pasmramtohub
                        cmp     cmd, #"U"       wz      ' ram to display
              if_z      jmp     #pasmramtodisplay
                        cmp     cmd, #"V"       wz      ' hub to display
              if_z      jmp     #pasmhubtodisplay           
                        cmp     cmd, #"E"       wz      ' convert 3 byte .raw format to 2 byte .ili format - hub to hub
              if_z      jmp     #rawtoiliformat
                        cmp     cmd, #"F"       wz      ' convert 3 byte .bmp format BGR to 2 byte ili format (same as E but order reversed)
              if_z      jmp     #bmptoiliformat              
 '                       cmp     cmd, #"W"       wz      ' lcdwritecom in pasm, not working
 '             if_z      jmp     #pasmlcdwritecom
                        cmp     cmd, #"X"       wz      ' merge icon and background based on a mask
              if_z      jmp     #mergeicons
                        cmp     cmd, #"Y"       wz      ' change the 137 output
              if_z      jmp     #changegroup
                        cmp     cmd, #"Z"       wz      ' set the 161 counters
              if_z      jmp     #set161          
                        cmp     cmd, #"I"       wz      ' init
              if_z      jmp     #init     
                        mov     err, cmd                ' error = cmd (unknown command)
                        jmp     #done
                        
' ----------------- common routines -------------------------------------

get_values              rdlong  hubaddr, hubptr         ' get hub address
                        rdlong  ramaddr, ramptr         ' get ram address
                        rdlong  len, lenptr             ' get length
                        mov     err, #5                 ' err=5
get_values_ret          ret

           ' Pass pasm_n = 0- 7 come to this with P0-P20 and P22 tristated and returns them as this too
set137                  or      dira,maskP22            ' pin 22 is an output
                        andn    outa,maskP22            ' set P22low so Y0-Y7 are all high
                        or      dira,maskP0P20          ' pins P0-P20 are outputs
                        and     outa,maskP0P2low        ' set these 3 pins low
                        or      outa,pasm_n             ' set the 137 pins
                        or      outa,maskP22            ' pin 22 high
set137_ret              ret                             ' return                        


load161pasm                                             ' uses ramaddr
                        or      outa,maskP0P20          ' set P0-P20 high     
                        or      dira,maskP0P20          ' output pins 0-20
                        mov     pasm_n,#0               ' group 0
                        call    #set137                 ' set the 137 output
                        and     outa,maskP0P18low       ' pins 0-18 set low
                        or      outa,ramaddr            ' output addres to 161 chips
                        or      outa,maskP19            ' clock high
                        or      outa,maskP20            ' load high
                        andn    outa,maskP19            ' clock low
                        andn    outa,maskP20            ' load low
                        or      outa,maskP19            ' clock high
                        or      outa,maskP20            ' load high 
load161pasm_ret         ret

stop                   jmp     #stop                  ' for debugging

memorytransfer          or      dira,maskP16P20         ' so /wr and other pins definitely high
                        or      outa,maskP16P20
                        mov     pasm_n,#1               ' back to group 1 for memory transfer
                        call    #set137                 ' as next routine will always be group 1
                        or      dira,maskP16P20         ' output pins 16-20
                        or      outa,maskP16P20         ' set P16-P20 high (P0-P15 set as inputs or outputs in the calling routine)
memorytransfer_ret      ret

busoutput               or      dira,maskP0P15          ' set prop pins 0-15 as outputs
busoutput_ret           ret

businput                and     dira,maskP16P31         ' set P0-P15 as inputs
businput_ret            ret  

delaynop                nop
                        nop
                        nop
                        nop
delaynop_ret            ret


' ------------------ single letter commands  -------------------------------------
 
' command S
pasmhubtoram            call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high
                        call    #busoutput              ' set prop pins P0-P15 as outputs
hubtoram_loop           and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP17            ' set mem write low
                        add     hubaddr,#2              ' increment by 2 bytes = 1 word. Put this here for small delay while writes
                        or      outa,maskP17            ' mem write high
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        djnz    len,#hubtoram_loop      ' loop this many times
                        jmp     #init                   ' tristate pins and listen for commands

' command T
pasmramtohub            call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high                             
                        call    #businput               ' set prop pins P0-P15 as inputs
                        andn    outa,maskP16            ' memory /rd low
ramtohub_loop           mov     data_16,ina             ' get the data
                        wrword  data_16,hubaddr         ' move data to hub
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        add     hubaddr,#2              ' increment the hub address 
                        djnz    len,#ramtohub_loop
                        or      outa,maskP16            ' memory /rd high  
                        or      dira,maskP0P15          ' %00000000_00000000_11111111_11111111 restore P0-P15as outputs
                        jmp     #init                   ' ' tristate pins and listen for commands

' command U
pasmramtodisplay        call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high                             
                        call    #businput               ' set prop pins 0-15 as inputs so doesn't interfere with the transfer
                        or      outa,maskP18            ' ILI_RS high
                        andn    outa,maskP16            ' memory /rd low  
ramtodisplay_loop       andn    outa,maskP20            ' ILI write low
                        or      outa,maskP20            ' ILI write high
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        djnz    len,#ramtodisplay_loop
                        or      outa,maskP16            ' memory /rd high  
                        or      dira,maskP0P15          ' %00000000_00000000_11111111_11111111 restore P0-P15as outputs
                        jmp     #init

' command V
pasmhubtodisplay        call    #get_values             ' get hubaddr,ramaddr,len
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high                             
                        call    #busoutput              ' set P0-P15 as outputs
hubtodisplay_loop       and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP20            ' ILI write low
                        or      outa,maskP20            ' ILI write high
                        add     hubaddr,#2              ' one word
                        djnz    len,#hubtodisplay_loop
                        jmp     #init

'command E
RawtoILIformat          ' takes a .raw 3 byte RRRRRRRR GGGGGGGG BBBBBBBB and converts to 2 byte RRRRRGGG GGGBBBBB
                        ' pass hubaddr, ramaddr and len
                        ' hubaddr is source location, len is number of pixels
                        ' ramaddr is destination in hub (messy naming) and length is 2/3 of blocklength
                        call    #get_values ' gets hubaddress, ramaddress and len (ignores ramaddress)
rawloop
                        rdbyte red,hubaddr
                        add hubaddr,#1
                        rdbyte green,hubaddr
                        add hubaddr,#1
                        rdbyte blue,hubaddr
                        add hubaddr,#1
                        call #rgbtoili
                        wrbyte ililow,ramaddr
                        add ramaddr,#1
                        wrbyte ilihigh,ramaddr
                        add ramaddr,#1
                        djnz    len,#rawloop            ' loop until done 
                        jmp     #init                   ' set pins to tristate

RGBtoILI                ' pass red,green, blue, returns ililow and ilihigh
                        shr     red,#3                  ' 000RRRRR 
                        shl     red,#3                  ' RRRRR000 
                        shr     green,#2                ' 00GGGGGG
                        mov     ilihigh,green           ' ilihigh = 00GGGGGG
                        shr     ilihigh,#3              ' ilihigh = 00000GGG
                        or      ilihigh,red             ' ilihigh = RRRRRGGG
                        and     green,#%00000111        ' 00000GGG
                        shl     green,#5                ' GGG00000
                        mov     ililow,green            ' ililow = GGG00000
                        shr     blue,#3                 ' blue = 000BBBBB
                        or      ililow,blue             ' ililow = GGGBBBBB
RGBtoILI_ret            ret

BMPtoILIformat          ' takes a .bmp 3 byte BBBBBBBB GGGGGGGG RRRRRRRR and converts to 2 byte RRRRRGGG GGGBBBBB
                        ' same as E above but BGR instead of RGB
                        ' pass hubaddr, ramaddr and len
                        ' hubaddr is source location, len is number of pixels
                        ' ramaddr is destination in hub (messy naming) and length is 2/3 of blocklength
                        call    #get_values ' gets hubaddress, ramaddress and len (ignores ramaddress)
bmploop
                        rdbyte blue,hubaddr
                        add hubaddr,#1
                        rdbyte green,hubaddr
                        add hubaddr,#1
                        rdbyte red,hubaddr
                        add hubaddr,#1
                        call #rgbtoili
                        wrbyte ililow,ramaddr
                        add ramaddr,#1
                        wrbyte ilihigh,ramaddr
                        add ramaddr,#1
                        djnz    len,#bmploop            ' loop until done 
                        jmp     #init                   ' set pins to tristate
' **** command X *********************

MergeIcons              call    #get_values ' gets hubaddress, ramaddress,len which are used here as background,icon,mask
                        mov     pasm_n,#59               ' do a single row
mergeiconsloop          rdbyte  ililow,len                 ' reuse ililow, so this is rdword mask,maskcounter
                        and     ililow,#%11111             ' mask off low 5 bits and use just the blue as this is a grayscale bitmap
                        rdword  red,hubaddr              ' reuse red, so actually this is rdword background,backgroundcounter                        
                        cmp     ililow,#%10000   wc       ' compare if >128 (ie mid level gray)
              if_c      jmp     #mergeskip
                        rdword  green,ramaddr            ' reuse green, so this is rdword iconpixel, iconpixelcounter 
                        wrword  green,hubaddr            ' if replace, then move icon pixel to the background     
mergeskip               add     hubaddr,#2
                        add     ramaddr,#2
                        add     len,#2
                        djnz    pasm_n,#mergeiconsloop            ' loop until done 
                        jmp     #init                   'set pins to tristate 

' *** command Y **********
changegroup             call    #get_values             'gets hubaddr, ramaddr,len which are used here as background,icon,mask           
                        mov     pasm_n,hubaddr          ' pass hubaddr
                        call    #set137                 ' change the group
                        jmp     #init

' *** command Z **********
set161                  call    #get_values             ' gets ramaddr = the 161 value to set
                        call    #load161pasm            ' ramaddr to the 161 chips
                        call    #memorytransfer         ' change to group 1
                        jmp     #init                        

                        

'pasmlcdwritecom         call    #get_values             ' use hubaddr as the data
'                        or      dira,maskP0P20          ' set these pins high (pass all pins tristated)
'                        or      outa,maskP0P20          '  set pins high
'                        mov     pasm_n,#2               '  mem transfer
'                        call    #set138                 ' set the 138
'                        andn    outa,maskP18            ' P18 ILIRS low
'                        and     outa,maskP16P31         ' set P0-P15 low
'                        or      outa,hubaddr            ' send out the data
'                        andn    outa,maskP17            ' ILI write low
'                        or      outa,maskP17            ' ILI write high
'                        jmp     #init                   ' set pins to tristate  

' variables
pasm_n                  long    0                                    ' general purpose value
data_16                 long    0                                    ' general purpose value
ililow                  long    0                                    ' low data byte 
ilihigh                 long    0                                    ' high data byte 
red                     long    0                                    ' red, green blue variables
green                   long    0
blue                    long    0           

' constants
Zero                    long    %00000000_00000000_00000000_00000000 ' used in several places
maskP0P2low             long    %11111111_11111111_11111111_11111000 ' P0-P2 low
maskP0P20               long    %00000000_00011111_11111111_11111111 ' P0-P18 enabled for output plus P19,P20    
maskP0P18low            long    %11111111_11111000_00000000_00000000 ' P0-P18 low
maskP16                 long    %00000000_00000001_00000000_00000000 ' pin 16
maskP17                 long    %00000000_00000010_00000000_00000000 ' pin 17
maskP18                 long    %00000000_00000100_00000000_00000000 ' pin 18
maskP19                 long    %00000000_00001000_00000000_00000000 ' pin 19
maskP20                 long    %00000000_00010000_00000000_00000000 ' pin 20
maskP22                 long    %00000000_01000000_00000000_00000000 ' pin 22
maskP16P31              long    %11111111_11111111_00000000_00000000 ' pin 16 to pin 31
maskP0P15               long    %00000000_00000000_11111111_11111111 ' for masking words
maskP16P20              long    %00000000_00011111_00000000_00000000
maskP0P20P22            long    %11111111_10100000_00000000_00000000 ' for returning all group pins HiZ
                        fit     496

CON
{{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// WAV-Player Program
//
// Author: Kwabena W. Agyeman
// Updated: 7/20/2011
// Designed For: P8X32A
// Version: 1.2
//
// Copyright (c) 2011 Kwabena W. Agyeman
// See end of file for terms of use.
//
// Update History:
//
// v1.0 - Original release - 7/6/2011.
// v1.1 - Tested with updated driver - 7/18/2011.
// v1.2 - Added support for dither mode - 7/20/2011.
//
// Plays a WAV file over and over again at different sample rates. The WAV file is specified by its file path name string that
// can be found in the code below. The WAV-Player program can play any standard WAV file. E.g. 16/8-Bit 16000/22050/44100-Hz.
//
// Nyamekye,
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}

CON
  _lpin = 21 ' -1 if unused.
  _rpin = 23 ' -1 if unused.

  _volume = 50 ' Default volume 0-100

  _ditherEnable = true ' "true" or "false" please.
  _ditherLevel = 4 ' 0 = Most Dither ... 31 = Least Dither.
  
VAR long spinPlayerStack[100]

PUB WavDemo
  DACEngineStart(constant(_lpin | (not(not(_ditherEnable)))), constant(_rpin | (not(not(_ditherEnable)))), _volume)
  if(_ditherEnable)
    DACDitherEngineStart(_lpin, _rpin, _ditherLevel) ' Never fails - no need to check return value.
  cognew(spinPlayer, @spinPlayerStack) ' Startup separate process.
  ' http://audio.online-convert.com/convert-to-wav      Online .mp3 to .wav converter
  'repeat ' Repeat forever.
  repeat until(wavePlayerState) ' Wait until start.
    

PUB spinPlayer ' Plays a WAV file over and over again.
  repeat ' Forever
    playWAVFile(string("piano.wav")) ' Supports WAV files up to ~2GB.

{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}

{{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// WAV-Player Digital to Analog Converter Engine
//
// Author: Kwabena W. Agyeman
// Updated: 7/20/2011
// Designed For: P8X32A
// Version: 1.3
//
// Copyright (c) 2011 Kwabena W. Agyeman
// See end of file for terms of use.
//
// Update History:
//
// v1.0 - Original release - 7/27/2010.
// v1.1 - Merged object with player - 7/6/2011.
// v1.2 - Fixed playback issue - 7/18/2011.
// v1.3 - Added dither cog - 7/20/2011.
//
// For each included copy of this object only one spin interpreter should access it at a time.
//
// Nyamekye,
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Audio Circuit:
//
//                  100OHM    1UF
// Left Pin Number  --R-------C-- Left Out (Use better audio circuit filtering for noise issues).
//                        |
//                        C 10NF
//                        |
//                       GND
//
//                  100OHM    1UF
// Right Pin Number --R-------C-- Right Out (Use better audio circuit filtering for noise issues). 
//                        |
//                        C 10NF
//                        |
//                       GND
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}

'OBJ  fat: "SD-MMC_FATEngine.spin"

VAR

  long leftValue, rightValue, samplesPerSecond, clocksPerSample, numberOfSamples, samplePosition, ditherLeft, ditherRight
  word dataBlock[512], callerPointer, callePointer, leftVolume, rightVolume, numberOfChannels, bitsPerSample 
  byte stopedOrStarted, beginOrEnd, cogIdentification, ditherCogIdentification

PUB leftChannelOutput '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the current sample going out to the left audio channel. The value returned is a 16 bit signed extended number.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return leftValue

PUB rightChannelOutput '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the current sample going out to the right audio channel. The value returned is a 16 bit signed extended number.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return rightValue

PUB fileSampleRateSetup(sampleRate) '' 4 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Setups the sample rate of the audio file. The samples are played at the sample rate.
'' //
'' // SampleRate - The new sample rate to playback sample at. Between 1 - 66,150 samples per second.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  clocksPerSample := (clkfreq / ((sampleRate <# 66_150) #> 1))

PUB fileSampleRate '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the sample rate of the audio file. The samples are played at the sample rate.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return samplesPerSecond

PUB fileSampleNumber '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the number of samples in the audio file. The samples are played at the sample rate.
'' //
'' // Valid while method "wavePlayerState" returns true.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return numberOfSamples

PUB fileSamplePosition '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the playback sample position in the audio file. The samples are played at the sample rate.
'' //
'' // Valid while method "wavePlayerState" returns true.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return (numberOfSamples - samplePosition)

PUB leftChannelVolume(newVolume) '' 4 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the volume of the left channel. (0 to 100).
'' //
'' // NewVolume - New volume to output at. Zero mutes the channel.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  leftVolume := (((newVolume <# 100) #> 0) * constant($1_00_00 / 100))

PUB rightChannelVolume(newVolume) '' 4 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the volume of the right channel. (0 to 100).
'' //
'' // NewVolume - New volume to output at. Zero mutes the channel.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  rightVolume := (((newVolume <# 100) #> 0) * constant($1_00_00 / 100))

PUB pauseWavePlayer(state) '' 4 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Pause or unpause the wave player. The method "wavePlayerState" has nothing to do with the wave player being paused.
'' //
'' // State - New wave player state. True for paused and false for unpaused.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  stopedOrStarted := (not(not(state)))

PUB wavePlayerState '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the wave player's state. True while playing a WAV file and false while not playing a WAV file.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  result or= beginOrEnd

PUB playWAVFile(filePathName) '' 56 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Plays a WAV file on the SD/MMC card - supports WAV files up to ~2GB and over ~1 second of audio.
'' //
'' // The WAV player supports 8-Bit 1/2 channel WAV files sampled between 1 Hz to 44,100 Hz.
'' //
'' // The WAV player supports 16-Bit 1/2 channel WAV files sampled between 1 Hz to 44,100 Hz.
'' //
'' // If an error occurs while reading the WAV file a string describing that error will be returned. Null otherwise.
'' //
'' // If an error occurs while reading the FAT file system a string describing that error will be returned. Null otherwise.
'' //
'' // FilePathName - A file system path string specifying the path of the WAV file to play.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  result := \playWAV(filePathName)
  \fat.unmountPartition

PRI playWAV(filePathName) | chunkSkip, blockAlign, byteRate, searchChunkSize ' 52 Stack Longs

  ifnot(fat.partitionMounted)
    fat.mountPartition(0)

  fat.openFile(filePathName, "R")

  if(fat.readLong <> $46_46_49_52)
    abort string("Not A RIFF File")

  if((fat.readLong + 8) <> fat.fileSize)
    abort string("RIFF Chunk Size Invalid")

  if(fat.readLong <> $45_56_41_57)
    abort string("Not A WAVE File")

  repeat while(fat.readLong <> $20_74_6D_66)
    searchChunkSize := fat.readLong

    if((searchChunkSize < 1) or (fat.fileSize == fat.fileTell))
      abort string("FMT Chunk Missing")

    fat.fileSeek(searchChunkSize + fat.fileTell)

  chunkSkip := (fat.readLong - 16)
  if(chunkSkip < 0)
    abort string("FMT Chunk Size Invalid")

  if(fat.readShort <> 1)
    abort string("Not A LPCM file")

  numberOfChannels := fat.readShort
  if((numberOfChannels < 1) or (2 < numberOfChannels))
    abort string("Unsupported Number Of Channels")

  samplesPerSecond := fat.readLong
  if((samplesPerSecond < 1) or (44_100 < samplesPerSecond))
    abort string("Unsupported Samples Per Second")

  clocksPerSample := (clkfreq / samplesPerSecond)
  byteRate := fat.readLong

  blockAlign := fat.readShort
  if((blockAlign <> 1) and (blockAlign <> 2) and (blockAlign <> 4))
    abort string("Unsupported Block Align")

  if(byteRate <> (samplesPerSecond * blockAlign))
    abort string("Invalid Byte Rate")

  bitsPerSample := fat.readShort
  if((bitsPerSample <> 8) and (bitsPerSample <> 16))
    abort string("Unsupported Bits Per Sample")

  fat.fileSeek(chunkSkip + fat.fileTell)

  repeat while(fat.readLong <> $61_74_61_64)
    searchChunkSize := fat.readLong

    if((searchChunkSize < 1) or (fat.fileSize == fat.fileTell))
      abort string("DATA Chunk Missing")

    fat.fileSeek(searchChunkSize + fat.fileTell)

  numberOfSamples := (fat.readLong / blockAlign)

  callerPointer := callePointer
  repeat while(callerPointer == callePointer) 
  fat.readData(@dataBlock[256 & callerPointer], 512)    
  samplePosition := numberOfSamples

  repeat (((((numberOfSamples * blockAlign) + 511) / 512) - 1) #> 0)
    not callerPointer
    repeat while(callerPointer == callePointer)
    fat.readData(@dataBlock[256 & callerPointer], 512)

PUB DACEngineStart(leftPinNumber, rightPinNumber, initialVolume) '' 9 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Starts up the DAC driver running on a cog.
'' //
'' // Returns true on success or false.
'' //
'' // LeftPinNumber - Pin to use to play the left channel audio. -1 to disable.
'' // RightPinNumber - Pin to use to play the right channel audio. -1 to disable.
'' // InitialVolume - Initial volume for the left and right audio channels. Between 0 and 100. 0 mutes.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  DACEngineStop
  if(chipver == 1)

    clocksPerSample := clkfreq
    rightVolume := leftVolume := (((initialVolume <# 100) #> 0) * constant($1_00_00 / 100))

    leftCounterSetup := ((leftPinNumber & $1F) + constant(%00110 << 26))
    rightCounterSetup := ((rightPinNumber & $1F) + constant(%00110 << 26))
    outputMask := (((leftPinNumber <> -1) & (|<leftPinNumber)) | ((rightPinNumber <> -1) & (|<rightPinNumber)))

    leftValueAddress := @leftValue
    rightValueAddress := @rightValue
    samplePositionAddress := @samplePosition
    dataBlockAddress := @dataBlock
    callePointerAddress := @callePointer
    numberOfChannelsAddress := @numberOfChannels
    bitsPerSampleAddress := @bitsPerSample
    leftVolumeAddress := @leftVolume
    rightVolumeAddress := @rightVolume
    stopedOrStartedAddress := @stopedOrStarted
    beginOrEndAddress := @beginOrEnd 
    ditherOutLeftAddress := @ditherLeft 
    ditherOutRightAddress := @ditherRight 
    
    cogIdentification := cognew(@initialization,  @clocksPerSample)
    result or= ++cogIdentification

PUB DACEngineStop '' 3 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Shuts down the DAC driver running on a cog.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if(cogIdentification)
    cogstop(-1 + cogIdentification~)

PUB DACDitherEngineStart(leftPinNumber, rightPinNumber, ditherLevel) '' 9 Stack Longs   

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Starts up the DAC dither driver running on a cog.
'' //
'' // The "DACEngineStart" method starts the regular non-dithering audio driver. If the audio quality is acceptable using it
'' // then the special dithering audio driver is not needed. If the audio quality is not acceptable then use the dither engine
'' // to eliminate both quantization noise from delta-modulation and pico-second jitter noise from on-chip crosstalk between
'' // nearby pins. Normally, these two noise sources generate distracting buzzes, whines, and hash, but dithering removes them 
'' // in exchange for lower-level white noise which does not draw attention. The dither level can be adjusted - the further 
'' // you get from COG0, the more dither you need. The special dither audio driver will use another cog in addition to the 
'' // cog used by the regular non-dithering audio driver. Both audio drivers can output to different audio pins to produce
'' // a dithered audio source and non-dithered audio source. To produce only a dithered audio source disable the left and
'' // right audio channel pins in the regular non-dithering audio driver by passing -1 as the left and right pin numbers. If
'' // only a non-dithered audio source is required do not call the "DACDitherEngineStart" method - this saves a cog.
'' //
'' // Returns true on success or false.
'' //
'' // LeftPinNumber - Pin to use to play the left channel audio. -1 to disable.
'' // RightPinNumber - Pin to use to play the right channel audio. -1 to disable.
'' // DitherLevel - Dither level for the left and right audio channels. 0 = Most Dither ... 31 = Least Dither.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  DACDitherEngineStop
  if(chipver == 1)

    ditherLeftCounterSetup := ((leftPinNumber & $1F) + constant(%00110 << 26))  
    ditherRightCounterSetup := ((rightPinNumber & $1F) + constant(%00110 << 26))  
    ditherOutputMask := (((leftPinNumber <> -1) & (|<leftPinNumber)) | ((rightPinNumber <> -1) & (|<rightPinNumber)))
    leftDitherShift := rightDitherShift := (ditherLevel & $1F)
    
    ditherLeftAddress := @ditherLeft
    ditherRightAddress := @ditherRight

    ditherCogIdentification := cognew(@ditherInitialization, 0)
    result or= ++ditherCogIdentification
  
PUB DACDitherEngineStop '' 3 Stack Longs 

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Shuts down the DAC dither driver running on a cog.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if(ditherCogIdentification)
    cogstop(-1 + ditherCogIdentification~)
  
PUB FATEngineStart(DOPin, CLKPin, DIPin, CSPin, WPPin, CDPin, RTCReserved1, RTCReserved2, RTCReserved3) '' 18 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Starts up the SDC driver running on a cog and checks out a lock for the driver.
'' //
'' // This method should only be called once for any number of included versions of this object.
'' //
'' // This method causes all included versions of this object to need re-mounting when called.
'' //
'' // Returns true on success or false.
'' //
'' // DOPin - The SPI data out pin from the SD card. Between 0 and 31.
'' // CLKPin - The SPI clock pin from the SD card. Between 0 and 31.
'' // DIPin - The SPI data in pin from the SD card. Between 0 and 31.
'' // CSPin - The SPI chip select pin from the SD card. Between 0 and 31.
'' // WPPin - The SPI write protect pin from the SD card holder. Between 0 and 31. -1 if not installed.
'' // CDPin - The SPI write protect pin from the SD card holder. Between 0 and 31. -1 if not installed.
'' // RTCReserved1 - Reserved parameter 1 for RTC compatible driver versions. Pass -1.
'' // RTCReserved2 - Reserved parameter 2 for RTC compatible driver versions. Pass -1.
'' // RTCReserved3 - Reserved parameter 3 for RTC compatible driver versions. Pass -1.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return fat.FATEngineStart(DOPin, CLKPin, DIPin, CSPin, WPPin, CDPin, RTCReserved1, RTCReserved2, RTCReserved3)

PUB FATEngineStop '' 6 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Shuts down the SDC driver running on a cog and returns the lock used by the driver.
'' //
'' // This method should only be called once for any number of included versions of this object.
'' //
'' // This method causes all included versions of this object to need re-mounting when called.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  fat.FATEngineStop

DAT

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       DAC Driver
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

                        org     0

' //////////////////////Initialization/////////////////////////////////////////////////////////////////////////////////////////

initialization          mov     ctra,               leftCounterSetup           ' Setup counter modes to duty cycle mode.
                        mov     ctrb,               rightCounterSetup          '
                        mov     frqa,               longAdjust                 '
                        mov     frqb,               longAdjust                 '
                        mov     dira,               outputMask                 '
                        mov     playerPointer,      dataBlockAddress           '

                        rdlong  timeCounter,        par                        ' Setup timing.
                        add     timeCounter,        cnt                        '

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Player                                                                                 
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

outerLoop               rdlong  buffer,             samplePositionAddress wz   ' Enable playback playback given samples. 
                        muxnz   playerMode,         #4                         '
                                   
                        rdword  buffer,             numberOfChannelsAddress    ' Setup number of channels.
                        cmp     buffer,             #1 wz                      '
                        muxz    playerMode,         #1                         '

                        rdword  buffer,             bitsPerSampleAddress       ' Setup bits per channel.
                        cmp     buffer,             #8 wz                      '
                        muxz    playerMode,         #2                         '

                        mov     counter,            #128                       ' Setup loop counter.
                        test    playerMode,         #1 wc                      '
if_c                    shl     counter,            #1                         '
                        test    playerMode,         #2 wc                      '
if_c                    shl     counter,            #1                         '

' //////////////////////Inner Loop/////////////////////////////////////////////////////////////////////////////////////////////

innerLoop               rdlong  buffer,             par                        ' Wait until next sample output period.
                        waitcnt timeCounter,        buffer                     '

                        rdbyte  buffer,             stopedOrStartedAddress wz  ' If stopped loop continously.
if_nz                   mov     frqa,               longAdjust                 '
if_nz                   mov     frqb,               longAdjust                 '
if_nz                   jmp     #innerLoop                                     '

                        movs    signedLongOutput,   #leftValueAddress          ' Get and output value.
                        movs    multiplicand,       #leftVolumeAddress         '
                        movs    unsignedLongOutput, #ditherOutLeftAddress      '
                        call    #decode                                        '
                        mov     frqa,               sampleBuffer               '

                        test    playerMode,         #1 wc                      ' Check number of channels.
                        test    playerMode,         #2 wz                      '
if_c_and_nz             sub     playerPointer,      #1                         '
if_c_and_z              sub     playerPointer,      #2                         '

                        movs    signedLongOutput,   #rightValueAddress         ' Get and output value.
                        movs    multiplicand,       #rightVolumeAddress        '
                        movs    unsignedLongOutput, #ditherOutRightAddress     ' 
                        call    #decode                                        '
                        mov     frqb,               sampleBuffer               '

                        rdlong  buffer,             samplePositionAddress      ' Decrement position.
                        cmpsub  buffer,             #1 wc, wz                  '
if_c                    wrlong  buffer,             samplePositionAddress      '

if_z                    andn    playerMode,         #4                         ' Disable playback given samples.    

                        test    playerMode,         #4 wc                      ' Playback begun.
if_c                    mov     buffer,             #$FF                       '
if_c_and_nz             wrbyte  buffer,             beginOrEndAddress          '

if_z                    wrbyte  buffer,             beginOrEndAddress          ' Playback be done.
   
                        djnz    counter,            #innerLoop                 ' Loop.

' //////////////////////Outer Loop/////////////////////////////////////////////////////////////////////////////////////////////

                        rdword  buffer,             callePointerAddress wz     ' Switch data block pointer.
                        sumz    buffer,             #1                         '
                        wrword  buffer,             callePointerAddress        '
if_nz                   mov     playerPointer,      dataBlockAddress           '

                        jmp     #outerLoop                                     ' Loop.

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Decode Value
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

decode                  test    playerMode,         #2 wc                      ' Read data depending on size.

if_c                    rdbyte  multiplyBuffer,     playerPointer              ' Read a byte or a word.   
if_c                    add     playerPointer,      #1                         '
if_nc                   rdword  multiplyBuffer,     playerPointer              '
if_nc                   add     playerPointer,      #2                         '

if_c                    sub     multiplyBuffer,     #128                       ' Sign extend.
if_c                    shl     multiplyBuffer,     #24                        '
if_nc                   shl     multiplyBuffer,     #16                        '
                        sar     multiplyBuffer,     #16                        '

                        test    playerMode,         #4 wz                      ' Zero playback given samples.  
if_z                    mov     multiplyBuffer,     #0                         '
                        
signedLongOutput        wrlong  multiplyBuffer,     0                          ' Update main memory.

multiplicand            rdword  multiplyCounter,    0                          ' Setup inputs.
                        mov     sampleBuffer,       #0                         '

                        abs     multiplyBuffer,     multiplyBuffer wc          ' Backup sign.
                        rcr     sampleBuffer,       #1 wz, nr                  '

multiplyLoop            shr     multiplyCounter,    #1 wc                      ' Preform multiplication.
if_c                    add     sampleBuffer,       multiplyBuffer             '
                        shl     multiplyBuffer,     #1                         '
                        tjnz    multiplyCounter,    #multiplyLoop              '

                        negnz   sampleBuffer,       sampleBuffer               ' Restore sign and center value.
unsignedLongOutput      wrlong  sampleBuffer,       0                          '                        
                        add     sampleBuffer,       longAdjust                 '

decode_ret              ret                                                    ' Return.

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Data
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

longAdjust              long    $80_00_00_00                                   ' Centers output.

' //////////////////////Configuration Settings/////////////////////////////////////////////////////////////////////////////////

leftCounterSetup        long    0
rightCounterSetup       long    0
outputMask              long    0

' //////////////////////Addresses//////////////////////////////////////////////////////////////////////////////////////////////

leftValueAddress        long    0
rightValueAddress       long    0
samplePositionAddress   long    0
dataBlockAddress        long    0
callePointerAddress     long    0
numberOfChannelsAddress long    0
bitsPerSampleAddress    long    0
leftVolumeAddress       long    0
rightVolumeAddress      long    0
stopedOrStartedAddress  long    0
beginOrEndAddress       long    0
ditherOutLeftAddress    long    0
ditherOutRightAddress   long    0 

' //////////////////////Run Time Variables/////////////////////////////////////////////////////////////////////////////////////

buffer                  res     1
counter                 res     1
playerPointer           res     1
playerMode              res     1
sampleBuffer            res     1
timeCounter             res     1
multiplyBuffer          res     1
multiplyCounter         res     1

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

                        fit     496

DAT

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       DAC Dither Driver
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

                        org     0

' //////////////////////Initialization/////////////////////////////////////////////////////////////////////////////////////////

ditherInitialization    mov     ctra,               ditherLeftCounterSetup     ' Setup counter modes to duty cycle mode.
                        mov     ctrb,               ditherRightCounterSetup    '

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Dither
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ditherLoop              test    leftLFSR,           leftTaps wc                ' Iterate left dither source
                        rcl     leftLFSR,           #1                         '

                        test    rightLFSR,          rightTaps wc               ' Iterate right dither source
                        rcl     rightLFSR,          #1                         '

                        rdlong  ditherLBuffer,      ditherLeftAddress          ' Read left source and shift dither.
                        mov     ditherLCounter,     leftLFSR                   '
                        sar     ditherLCounter,     leftDitherShift            '

                        rdlong  ditherRBuffer,      ditherRightAddress         ' Read right source and shift dither.    
                        mov     ditherRCounter,     rightLFSR                  '
                        sar     ditherRCounter,     rightDitherShift           '

                        add     ditherLBuffer,      ditherLCounter             ' Apply dither.
                        add     ditherLBuffer,      ditherAdjust               '
                        
                        mov     frqa,               ditherLBuffer              ' Output.
                                               
                        add     ditherRBuffer,      ditherRCounter             ' Apply dither.
                        add     ditherRBuffer,      ditherAdjust               '
                        
                        mov     frqb,               ditherRBuffer              ' Output. 

                        mov     dira,               ditherOutputMask           ' Repeat.
                        jmp     #ditherLoop                                    ' 
                        
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'                       Data
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

ditherAdjust            long    $80_00_00_00                                   ' Prevents poping. 

leftTaps                long    $A4_00_00_80                                   ' Left LFSR taps.
rightTaps               long    $80_A0_10_00                                   ' Right LFSR taps.

leftLFSR                long    1                                              ' Initial value.
rightLFSR               long    1                                              ' Initial value.

' //////////////////////Configuration Settings/////////////////////////////////////////////////////////////////////////////////

ditherLeftCounterSetup  long    0
ditherRightCounterSetup long    0
ditherOutputMask        long    0

leftDitherShift         long    0
rightDitherShift        long    0

' //////////////////////Addresses//////////////////////////////////////////////////////////////////////////////////////////////

ditherLeftAddress       long    0
ditherRightAddress      long    0

' //////////////////////Run Time Variables/////////////////////////////////////////////////////////////////////////////////////

ditherLBuffer           res     1
ditherLCounter          res     1

ditherRBuffer           res     1
ditherRCounter          res     1

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

                        fit     496

{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}                   